from typing import Dict, Iterable, List, Optional, Tuple, Union

import attr
import lmfit
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import linregress


from scipy.ndimage import  map_coordinates, uniform_filter1d
from scipy.interpolate import RegularGridInterpolator
import proplot

from skultrafast import dv, plot_helpers
from skultrafast.dataset import TimeResSpec


def inbetween(a, lower, upper):
    return np.logical_and(a >= lower, a <= upper)


@attr.s(auto_attribs=True)
class CLSResult:
    wt: List[float]
    slopes: List[float]
    intercepts: List[float]
    intercept_errors: Optional[List[float]] = None
    slope_errors: Optional[List[float]] = None
    lines: Optional[np.ndarray] = None
    exp_fit_result_: Optional[lmfit.model.ModelResult] = None

    def exp_fit(self, start_taus: List[float], use_const=True, use_weights=True):
        mod = lmfit.models.ConstantModel()
        vals = {}

        for i, v in enumerate(start_taus):
            prefix = 'abcdefg'[i] + '_'
            mod += lmfit.models.ExponentialModel(prefix=prefix)
            vals[prefix + 'decay'] = v
            vals[prefix + 'amplitude'] = 0.5 / len(start_taus)

        for p in mod.param_names:
            if p.endswith('decay'):
                mod.set_param_hint(p, min=0)
            if p[:3] == 'amp':
                mod.set_param_hint(p, min=0)
        #mod.set_param_hint('c', min=-0)
        if use_const:
            c = max(np.min(self.slopes), 0)
        else:
            c = 0
            mod.set_param_hint('c', vary=False)

        if self.slope_errors is not None and use_weights:
            weights = 1 / np.array(self.slope_errors)
        else:
            weights = None
        res = mod.fit(self.slopes,
                      weights=weights,
                      x=self.wt,
                      c=c,
                      **vals)
        self.exp_fit_result_ = res
        return res


    def plot_cls(self, ax=None, model_style: Dict=None, symlog=False, **kwargs):
        if ax is None:
            ax = plt.gca()
        ec = ax.errorbar(self.wt, self.slopes, self.slope_errors, **kwargs)
        if symlog:
            ax.set_xscale('symlog', linthreshx=1)
        plot_helpers.lbl_trans(ax=ax, use_symlog=symlog)

        ax.set(xlabel=plot_helpers.time_label, ylabel='Slope')
        m_line = None
        if self.exp_fit_result_:
            xu = np.linspace(min(self.wt), max(self.wt), 300)
            yu = self.exp_fit_result_.eval(x=xu)
            style = dict(c='k', zorder=1.8)
            if model_style:
                style.update(model_style)
            ax.plot(xu, yu, color='k', zorder=1.8)
        return ec, m_line


@attr.s(auto_attribs=True)
class TwoDimPlotter:
    ds: 'TwoDim' = attr.ib()

    def contour(self, *times, region=None, ax=None, subplots_kws={}, aspect=None,
                direction='vertical', scale="firstmax", average=None):
        ds = self.ds
        idx = [ds.t_idx(i) for i in times]
        if ax is None:
            if aspect is None:
                aspect = ds.probe_wn.ptp() / ds.pump_wn.ptp()
            if direction[:1] == 'v':
                nrows = len(idx)
                ncols = 1
            else:
                nrows = 1
                ncols = len(idx)
            fig, ax = proplot.subplots(nrows=nrows, ncols=ncols, **subplots_kws,
                             aspect=aspect)
        if scale == 'fullmax':
            m = abs(ds.spec2d).max()
        elif scale == 'firstmax':
            m = abs(ds.spec2d[idx[0], ...]).max()
        else:
            raise ValueError("scale must be either fullmax or firstmax")

        if average is not None:
            s2d = uniform_filter1d(ds.spec2d, average, 0, mode="nearest")
        else:
            s2d = ds.spec2d

        for i, k in enumerate(idx):
            c = ax[i].contourf(ds.probe_wn,
                               ds.pump_wn,
                               s2d[k].T,
                               N=np.linspace(-m, m, 21),
                               cmap='Div',
                               linewidths=[0.7],
                               linestyles=['-'])
            start = max(ds.probe_wn.min(), ds.pump_wn.min())
            ax[i].axline((start, start), slope=1, c='k', lw=0.5)
            if ds.t[k] < 1:
                title = '%.0f fs' % (ds.t[k] * 1000)
            else:
                title = '%.1f ps' % (ds.t[k])
            ax[i].format(title=title, titleloc='ul', titleweight='bold')

        ax.format(xlabel='Probe Freq. / cm$^{-1}$',
                  ylabel='Pump Freq. / cm$^{-1}$',
                  xlim=region,
                  ylim=region)
        return c, ax

    def movie_contour(self, fname, contour_kw={}, subplots_kw={}):
        from matplotlib.animation import FuncAnimation

        c, ax = self.contour(self.ds.t[0], **subplots_kw)
        fig = ax.get_figure()
        frames = self.ds.t
        std_kws = {}
        def func(x):
            ax.cla()
            self.contour(x, ax=ax, scale="fullmax", **subplots_kw)
        ani = FuncAnimation(fig=fig, func=func, frames=frames)
        ani.save(fname)

    def plot_cls(self):
        pass

    def pump_slice_amps(self, ax=None):
        if ax is None:
            ax = plt.gca()
        ax.plot(self.ds.pump_wn, self.ds.pump_slice_amp())
        plot_helpers.ir_mode()
        ax.set(xlabel=plot_helpers.freq_label, ylabel='Slice Amp. [mOD]')


    def elp(self, spec_i, offset=None, p=None):
        fig, (ax, ax1) = plt.subplots(2, figsize=(3, 6), sharex='col')
        d = self.spec2d[spec_i].real.copy()[::, ::].T
        interpol = RegularGridInterpolator((self.pump_wn, self.probe_wn,), d[::, ::], bounds_error=False)
        m = abs(d).max()
        ax.pcolormesh(self.probe_wn, self.pump_wn, d, cmap='seismic', vmin=-m, vmax=m)

        ax.set(ylim=(self.pump_wn.min(), self.pump_wn.max()), xlim=(self.probe_wn.min(), self.probe_wn.max()))
        ax.set_aspect(1)
        if offset is None:
            offset = self.pump_wn[np.argmin(np.min(d, 1))] - self.probe_wn[np.argmin(np.min(d, 0))]
        if p is None:
            p = self.probe_wn[np.argmin(np.min(d, 0))]
        y_diag = self.probe_wn + offset
        y_antidiag = -self.probe_wn + 2 * p + offset
        ax.plot(self.probe_wn, y_diag, lw=1)
        ax.plot(self.probe_wn, y_antidiag, lw=1)

        diag = interpol(np.column_stack((y_diag, self.probe_wn)))
        antidiag = interpol(np.column_stack((y_antidiag, self.probe_wn)))
        ax1.plot(self.probe_wn, diag)
        ax1.plot(self.probe_wn, antidiag)
        return


@attr.s(auto_attribs=True)
class TwoDim:
    t: np.ndarray
    "Array of the waiting times"
    pump_wn: np.ndarray
    "Array with the pump-wavenumbers"
    probe_wn: np.ndarray
    "Array with the probe-wavenumbers"
    spec2d: np.ndarray
    "Array with the data"
    info: Dict = {}
    "Meta Info"
    cls_result_: Optional[CLSResult] = None
    "Contains the data from a CLS analysis"
    plot: 'TwoDimPlotter' = attr.Factory(TwoDimPlotter, True) # typing: Ignore
    "Plot object offering plotting methods"
    interpolator_: Optional[RegularGridInterpolator] = None # typing: Ignore

    def _make_int(self, ):
        RegularGridInterpolator((self.t, self.probe_wn, self.pump_wn), self.spec2d, bounds_error=False)

    def __attrs_post_init__(self):
        n, m, k = self.t.size, self.probe_wn.size, self.pump_wn.size
        if self.spec2d.shape != (n, m, k):
            raise ValueError("Data shape not equal to t, wn_probe, wn_pump shape"
                             f"{self.spec2d.shape} != {n, m, k}")

        self.spec2d = self.spec2d.copy()
        self.probe_wn = self.probe_wn.copy()
        self.pump_wn = self.pump_wn.copy()

        i1 = np.argsort(self.pump_wn)
        self.pump_wn = self.pump_wn[i1]
        i2 = np.argsort(self.probe_wn)
        self.probe_wn = self.probe_wn[i2]
        self.spec2d = self.spec2d[:, :, i1][:, i2, :]

    def copy(self) -> 'TwoDim':
        cpy = attr.evolve(self)
        cpy.plot = TwoDimPlotter(cpy)  # typing: ignore
        return cpy

    def t_idx(self, t: Union[float, Iterable[float]]) -> int:
        "Return nearest idx to nearest time value"
        return dv.fi(self.t, t)

    def probe_idx(self, wn: Union[float, Iterable[float]]) -> int:
        "Return nearest idx to nearest probe_wn value"
        return dv.fi(self.probe_wn, wn)

    def pump_idx(self, wn: Union[float, Iterable[float]]) -> int:
        "Return nearest idx to nearest pump_wn value"
        return dv.fi(self.pump_wn, wn)

    def select_range(self, pump_range, probe_range, invert=False) -> 'TwoDim':
        """
        Return a dataset containing only the selected region.
        """
        pu_idx = inbetween(self.pump_wn, min(pump_range), max(pump_range))
        pr_idx = inbetween(self.probe_wn, min(probe_range), max(probe_range))

        if invert:
            pu_idx = not pu_idx
            pr_idx = not pr_idx

        ds = self.copy()
        ds.spec2d = ds.spec2d[:, pr_idx, :][:, :, pu_idx]
        ds.pump_wn = ds.pump_wn[pu_idx]
        ds.probe_wn = ds.probe_wn[pr_idx]
        return ds

    def select_t_range(self, t_min: float= -np.inf, t_max: float=np.inf) -> 'TwoDim':
        """"
        Returns a dataset only containing the data within given time limits.
        """
        idx = inbetween(self.t, t_min, t_max)
        ds = self.copy()
        ds.t = ds.t[idx]
        ds.spec2d = ds.spec2d[idx, :, :]
        return ds

    def integrate_pump(self, lower: float = -np.inf, upper: float = np.inf) -> TimeResSpec:
        """
        Calculate and return 1D Time-resolved spectra for given range.

        Parameters
        ----------
        lower : float
            Lower pump wl
        upper : float
            upper pump wl

        Returns
        -------
        TimeResSpec
            The corresponding 1D Dataset
        """
        pu_idx = inbetween(self.pump_wn, lower, upper)
        data = np.trapz(self.spec2d[:, :, pu_idx], self.pump_wn[pu_idx], axis=-1)
        return TimeResSpec(self.probe_wn, self.t, data, freq_unit='cm')

    def single_cls(self, t, pr_range=9, pu_range=7, mode='neg') -> Tuple[np.ndarray, np.ndarray, object]:
        """
        Calculate the CLS for single 2D spectrum.

        Parameters
        ----------
        t : float
            Delay time of the spectrum to analyse
        pr_range : float, optional
            How many wavenumbers away from the maximum to use for
            determining the exact position, by default 9
        pu_range : float, optional
            The range around the pump-maxima used for calculating
            the CLS.
        mode : str, optional
            negative or positive maximum, by default 'neg'

        Returns
        -------
        (x, y, r)
            Return x, y and the regression result r
        """
        pu = self.pump_wn
        pr = self.probe_wn
        spec = self.spec2d[self.t_idx(t), :, :].T
        if mode == 'pos':
            spec = -spec
        pu_max = pu[np.argmin(np.min(spec, 1))]
        pu_idx = (pu < pu_max + pu_range) & (pu > pu_max - pu_range)
        spec = spec[pu_idx, :]
        l = []

        for s in spec:
            m = np.argmin(s)
            pr_max = pr[m]
            i = u_idx = (pr < pr_max + pr_range) & (pr > pr_max - pr_range)
            cen_of_m = np.average(pr[i], weights=s[i])
            l.append(cen_of_m)

        x = pu[pu_idx]
        y = np.array(l)
        r = linregress(x, y)
        return x, y, r

    def cls(self, **cls_args):
        slopes, slope_errs = [], []
        lines = []
        intercept = []
        intercept_errs = []
        for d in self.t:
            x, y, r = self.single_cls(d, **cls_args)
            slopes.append(r.slope)
            slope_errs.append(r.stderr)
            lines.append(np.column_stack((x, y)))
            intercept.append(r.intercept)
            intercept_errs.append(r.intercept_stderr)
        res = CLSResult(self.t, slopes=slopes, slope_errors=slope_errs,
                        lines=lines, intercepts=intercept, intercept_errors=intercept_errs)
        self.cls_result_ = res
        return res

    def extract_line_spec(self, offset):
        if not self.interpolator_:
            self.interpolator_ = self._make_int()

    def bin_t(self, n: int):
        pass


    #def pump_slice_amp(sel